package web.multitask.trismegistoservices.api;


import lombok.AllArgsConstructor;
import org.json.JSONObject;
import org.springframework.http.ResponseEntity;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.web.bind.annotation.*;
import web.multitask.trismegistoservices.model.User;
import web.multitask.trismegistoservices.repository.UserRespository;
import web.multitask.trismegistoservices.utils.JWTokenUtil;

import java.math.BigInteger;
import java.util.Objects;

import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;


@RestController
@RequestMapping("/token")
@CrossOrigin
@AllArgsConstructor
class JWTokenApi {

    private final JWTokenUtil jwtTokenUtil;
    private final UserRespository userRepo;

    @PostMapping("/authenticate")
    public ResponseEntity<?> createAuthenticationToken(@RequestBody String authenticationRequest) {
        JSONObject response;
        JSONObject json = new JSONObject(authenticationRequest);
        String username = json.getString("username");
        try {
            UserDetails userDetails = userRepo.findByUsername(username);
            if (!Objects.equals(userDetails.getPassword(), json.getString("password"))) {
                response = new JSONObject().put("token", "").put("message", "Invalid Credentials").put("status", false);
                return ResponseEntity.status(401).body(response.toMap());
            } else {
                return ResponseEntity.ok(new JSONObject().put("token", jwtTokenUtil.generateToken((User) userDetails, json.optBigInteger("ms", BigInteger.valueOf(3600000)))).put("message", "Generated").put("status", true).toMap());
            }
        } catch (Exception e) {
            response = new JSONObject().put("token", "").put("message", "Invalid Credentials").put("status", false);
            return ResponseEntity.status(401).body(response.toMap());
        }
    }

    @PostMapping("/validate")
    public ResponseEntity<?> validateToken(@RequestBody String token) {
        JSONObject json = new JSONObject(token);
        JSONObject response = new JSONObject();

        if (!jwtTokenUtil.validateToken(json.getString("token"))) {
            response.put("message", "Invalid Token").put("status", false);
            return ResponseEntity.status(401).body(response.toMap());
        }

        String dataToken = jwtTokenUtil.getDataToken(json.getString("token"));
        if (dataToken == null) {
            response.put("message", "Invalid Token").put("status", false);
            return ResponseEntity.status(401).body(response.toMap());
        }

        boolean isTokenExpired = jwtTokenUtil.isTokenExpired(json.getString("token"));
        if (isTokenExpired) {
            response.put("message", "Expired Token").put("status", false);
            return ResponseEntity.status(403).body(response.toMap());
        }

        try {
            UserDetails userDetails = userRepo.findByUsername(new JSONObject(dataToken).getString("username"));
            if (userDetails == null) {
                response.put("message", "Invalid Token").put("status", false);
                return ResponseEntity.status(401).body(response.toMap());
            }
        } catch (Exception e) {
            response.put("message", "Invalid Token").put("status", false);
            return ResponseEntity.status(401).body(response.toMap());
        }

        response.put("message", "Valid Token").put("status", true);
        return ResponseEntity.ok(response.toMap());
    }

    @PostMapping("/service/authenticate")
    public ResponseEntity<?> generateToken(@RequestBody String token) {
        JSONObject json = new JSONObject(token);
        UserDetails userDetails = userRepo.findByUsername(json.getString("username"));
        if (userDetails == null) {
            return ResponseEntity.status(401).body(new JSONObject().put("token", "").put("message", "Invalid Credentials").put("status", false).toMap());
        } else if (userDetails.getAuthorities().stream().anyMatch(a -> a.getAuthority().equals("SERVICE"))) {
            return ResponseEntity.ok(new JSONObject().put("token", jwtTokenUtil.generateToken((User) userDetails, json.optBigInteger("ms", BigInteger.valueOf(3600000)))).put("message", "Generated").put("status", true).toMap());
        } else {
            return ResponseEntity.status(401).body(new JSONObject().put("token", "").put("message", "Invalid Credentials").put("status", false).toMap());
        }
    }

    @PostMapping("/remaining")
    public ResponseEntity<?> remainingTime(@RequestBody String token) {
        JSONObject json = new JSONObject(token);
        try{
            int remaining = jwtTokenUtil.getExperyTime(json.getString("token"));
            return ResponseEntity.ok(new JSONObject().put("remaining", remaining).put("message", "OK").put("status", true).toMap());
        } catch (Exception e) {
            return ResponseEntity.status(401).body(new JSONObject().put("remaining", 0).put("message", "Invalid Token").put("status", false).toMap());
        }
    }

}