package web.multitask.trismegistoservices.services;

import java.awt.Color;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.IntStream;

import org.apache.poi.ss.usermodel.*;
import org.apache.poi.xssf.usermodel.XSSFCell;
import org.apache.poi.xssf.usermodel.XSSFCellStyle;
import org.apache.poi.xssf.usermodel.XSSFColor;
import org.apache.poi.xssf.usermodel.XSSFRow;
import org.apache.poi.xssf.usermodel.XSSFSheet;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.apache.poi.xssf.usermodel.extensions.XSSFCellBorder.BorderSide;
import org.json.JSONArray;
import org.json.JSONObject;
import org.springframework.core.io.ByteArrayResource;
import org.springframework.core.io.Resource;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;

@Service
public class ExcelService  {

    public byte[] generateExcel(JSONObject json,int size) throws IOException {
        ByteArrayOutputStream outByteStream = new ByteArrayOutputStream();
        XSSFWorkbook workbook = new XSSFWorkbook();
        createSheet(workbook, json,size);
        try {
            workbook.write(outByteStream);
            workbook.close();
            return outByteStream.toByteArray();
        } catch (Exception e) {
            workbook.close();
            System.out.println(e.getMessage());
            return null;
        }
    }

    public void createSheet(XSSFWorkbook workbook, JSONObject json, int sizeOfBytes) {
        AtomicReference<List<CompletableFuture<Boolean>>> results = new AtomicReference<>(new ArrayList<>());
        if (json.optJSONArray("sheetArray", new JSONArray()).isEmpty()) {
            XSSFSheet sheet = workbook.createSheet(json.optString("sheet_name", "no_name"));
            buildSheet(workbook, sheet, json);
            for (int i = 0; i < json.getJSONArray("identifiers").length(); i++) {
                sheet.setColumnWidth(i, 16 * 512);
            }
        } else {
            JSONArray sheetArray = json.optJSONArray("sheetArray", new JSONArray());
            if (sizeOfBytes < 100000) {
                for (int i = 0; i < sheetArray.length(); i++) {
                    JSONObject jsonObject = sheetArray.getJSONObject(i);
                    XSSFSheet sheet = workbook.createSheet(jsonObject.optString("sheet_name", "no_name"));
                    JSONArray identifiers = jsonObject.getJSONArray("identifiers");
                    for (int j = 0; j < identifiers.length(); j++) {
                        sheet.setColumnWidth(j, 16 * 512);
                    }
                    buildSheet(workbook, sheet, jsonObject);
                }
            } else {
                XSSFSheet[] sheets = new XSSFSheet[sheetArray.length()];
                for (int i = 0; i < sheetArray.length(); i++) {
                    JSONObject jsonObject = sheetArray.getJSONObject(i);
                    sheets[i] = workbook.createSheet(jsonObject.optString("sheet_name", "no_name"));
                }

                IntStream.range(0, sheetArray.length()).parallel().forEach(i -> {
                    JSONObject jsonObject = sheetArray.getJSONObject(i);
                    JSONArray identifiers = jsonObject.getJSONArray("identifiers");
                    for (int j = 0; j < identifiers.length(); j++) {
                        sheets[i].setColumnWidth(j, 16 * 512);
                    }
                    CompletableFuture<Boolean> result = buildSheetAsync(workbook, sheets[i], jsonObject);
                    results.get().add(result);
                });
//                sheetArray.toList().stream().parallel().forEach(sheetObject -> {
//                    HashMap<String, Object> sheetMap = (HashMap<String, Object>) sheetObject;
//                    JSONObject jsonObject = new JSONObject(sheetMap);
//                    JSONArray identifiers = jsonObject.getJSONArray("identifiers");
//                    for (int j = 0; j < identifiers.length(); j++) {
//                        sheets[i].setColumnWidth(j, 16 * 512);
//                    }
//                    CompletableFuture<Boolean> result = buildSheetAsync(workbook, sheets[i], jsonObject);
//                    results.get().add(result);
//                });

                CompletableFuture.allOf(results.get().toArray(new CompletableFuture[0])).join();
            }
        }
    }

    public void buildSheet(XSSFWorkbook workbook, XSSFSheet sheet, JSONObject json) {

        final XSSFCellStyle defaultStyle = createCellStyle(workbook, new JSONObject());
        List<String> stylesSaved = new ArrayList<>();
        List<XSSFCellStyle> styleObjectsSaved = new ArrayList<>();

        XSSFRow row;
        XSSFCell cell;
        XSSFCellStyle cellStyle;

        JSONArray identifiers = json.optJSONArray("identifiers", new JSONArray());
        JSONArray data = json.optJSONArray("data", new JSONArray());
        JSONArray styles = json.optJSONArray("styles", new JSONArray());

        fillStylesArray(styles,workbook,stylesSaved,styleObjectsSaved);

        for (int i = 0; i < data.length(); i++) {
            row = sheet.createRow(i);
            JSONObject row_data = data.optJSONObject(i);
            for (int j = 0; j < identifiers.length(); j++) {
                cell = row.createCell(j);
                JSONArray styleArray = styles.optJSONArray(i, new JSONArray());
                JSONObject style = styleArray.optJSONObject(j, new JSONObject());
                if(style.isEmpty()){
                    cell.setCellStyle(defaultStyle);
                }else{
                    if(!stylesSaved.contains(style.toString())){
                        stylesSaved.add(style.toString());
                        cellStyle = createCellStyle(workbook,style);
                        cell.setCellStyle(cellStyle);
                        styleObjectsSaved.add(cellStyle);
                    }else{
                        cell.setCellStyle(styleObjectsSaved.get(stylesSaved.indexOf(style.toString())));
                    }
                }
                cell.setCellValue(row_data.optString(identifiers.optString(j, "")));
            }
        }

    }

    @Async
    public CompletableFuture<Boolean> buildSheetAsync(XSSFWorkbook workbook, XSSFSheet sheet, JSONObject json) {

        final XSSFCellStyle defaultStyle = createCellStyle(workbook, new JSONObject());
        List<String> stylesSaved = new ArrayList<>();
        List<XSSFCellStyle> styleObjectsSaved = new ArrayList<>();

        XSSFRow row;
        XSSFCell cell;
        XSSFCellStyle cellStyle;

        JSONArray identifiers = json.optJSONArray("identifiers", new JSONArray());
        JSONArray data = json.optJSONArray("data", new JSONArray());
        JSONArray styles = json.optJSONArray("styles", new JSONArray());

       fillStylesArray(styles,workbook,stylesSaved,styleObjectsSaved);

        for (int i = 0; i < data.length(); i++) {
             row = sheet.createRow(i);
            JSONObject row_data = data.optJSONObject(i);
            for (int j = 0; j < identifiers.length(); j++) {
                 cell = row.createCell(j);
                JSONArray styleArray = styles.optJSONArray(i, new JSONArray());
                JSONObject style = styleArray.optJSONObject(j, new JSONObject());
                if(style.isEmpty()){
                    cell.setCellStyle(defaultStyle);
                }else{
                    if(!stylesSaved.contains(style.toString())){
                        stylesSaved.add(style.toString());
                         cellStyle = createCellStyle(workbook,style);
                        cell.setCellStyle(cellStyle);
                        styleObjectsSaved.add(cellStyle);
                    }else{
                        cell.setCellStyle(styleObjectsSaved.get(stylesSaved.indexOf(style.toString())));
                    }
                }
                cell.setCellValue(row_data.optString(identifiers.optString(j, "")));
            }
        }

        return CompletableFuture.completedFuture(true);
    }

    private XSSFCellStyle createCellStyle(XSSFWorkbook workbook, JSONObject styleJson) {

        XSSFColor background = getXSSFColor(styleJson.optString("background", "#ffffff"));
        XSSFColor foreground = getXSSFColor(styleJson.optString("foreground", "#333333"));
        boolean bold = styleJson.optBoolean("bold", false);
        boolean border = styleJson.optBoolean("border", true);
        String align = styleJson.optString("textAlign", "center");
        XSSFCellStyle style = workbook.createCellStyle();
        Font font = workbook.createFont();
        style.setAlignment( align.equals("center") ? HorizontalAlignment.CENTER : align.equals("left") ? HorizontalAlignment.LEFT : HorizontalAlignment.RIGHT);
        style.setWrapText(styleJson.optBoolean("wrapText", false));
        style.setVerticalAlignment(VerticalAlignment.CENTER);
        style.setFillPattern(FillPatternType.SOLID_FOREGROUND);
        style.setBorderBottom(BorderStyle.THIN);
        style.setBorderTop(BorderStyle.THIN);
        style.setBorderLeft(BorderStyle.THIN);
        style.setBorderRight(BorderStyle.THIN);

        if (border) {
            style.setBorderColor(BorderSide.BOTTOM, getXSSFColor("#333333"));
            style.setBorderColor(BorderSide.TOP, getXSSFColor("#333333"));
            style.setBorderColor(BorderSide.LEFT, getXSSFColor("#333333"));
            style.setBorderColor(BorderSide.RIGHT, getXSSFColor("#333333"));
        }else{
            style.setBorderColor(BorderSide.BOTTOM, getXSSFColor("#aaaaaa"));
            style.setBorderColor(BorderSide.TOP, getXSSFColor("#aaaaaa"));
            style.setBorderColor(BorderSide.LEFT, getXSSFColor("#aaaaaa"));
            style.setBorderColor(BorderSide.RIGHT, getXSSFColor("#aaaaaa"));
        }

        font.setBold(bold);
        style.setFont(font);

        if (background != null) {
            style.setFillForegroundColor(background);
        }

        if (foreground != null) {
            style.getFont().setColor(foreground);
        }

        return style;
    }

    public XSSFColor getXSSFColor(String rgb) {
        int red = Integer.parseInt(rgb.substring(1, 3), 16);
        int green = Integer.parseInt(rgb.substring(3, 5), 16);
        int blue = Integer.parseInt(rgb.substring(5, 7), 16);
        Color awtColor = new Color(red, green, blue);
        XSSFColor xssfColor = new XSSFColor();
        xssfColor
                .setRGB(new byte[] { (byte) awtColor.getRed(), (byte) awtColor.getGreen(), (byte) awtColor.getBlue() });
        return xssfColor;
    }

    public Resource convertByteToResource(byte[] data, String filename) {
        return new ByteArrayResource(data) {
            @Override
            public String getFilename() {
                return filename;
            }
        };
    }

    public void fillStylesArray(JSONArray styles, XSSFWorkbook workbook, List<String> stylesSaved, List<XSSFCellStyle> styleObjectsSaved) {
        for (int i = 0; i < styles.length(); i++) {
            JSONArray styleArray = styles.optJSONArray(i, new JSONArray());
            for (int j = 0; j < styleArray.length(); j++) {
                JSONObject style = styleArray.optJSONObject(j, new JSONObject());
                if(!stylesSaved.contains(style.toString())){
                    stylesSaved.add(style.toString());
                    XSSFCellStyle cellStyle = createCellStyle(workbook,style);
                    styleObjectsSaved.add(cellStyle);
                }
            }
        }
    }
}